import argparse
import sys
from unittest import TestCase

from yautil import Subcommand, SubcommandParser


class TestSubcommand(TestCase):

    def test_basic(self):
        cmd = ''

        class CmdA(Subcommand):
            def on_parser_init(self, parser: argparse.ArgumentParser):
                pass

            def on_command(self, args):
                assert cmd == 'cmda'

        parser = SubcommandParser()

        parser.add_subcommands(CmdA())

        cmd = 'cmda'
        args = parser.parse_args([cmd])
        parser.exec_subcommands(args)

    def test_subcommand_naming(self):
        cmd = ''

        class CmdA(Subcommand):
            def on_parser_init(self, parser: argparse.ArgumentParser):
                pass

            def on_command(self, args):
                assert cmd == 'X'

        parser = SubcommandParser()

        parser.add_subcommands(CmdA(name='X'))

        cmd = 'X'
        args = parser.parse_args([cmd])
        parser.exec_subcommands(args)

    def test_multiple_subcommand_regs(self):
        cmd = ''

        class CmdA(Subcommand):
            def on_parser_init(self, parser: argparse.ArgumentParser):
                pass

            def on_command(self, args):
                assert cmd == 'cmda'

        class CmdB(Subcommand):
            def on_parser_init(self, parser: argparse.ArgumentParser):
                pass

            def on_command(self, args):
                assert cmd == 'cmdb'

        class CmdC(Subcommand):
            def on_parser_init(self, parser: argparse.ArgumentParser):
                pass

            def on_command(self, args):
                assert cmd == 'cmdc'

        parser = SubcommandParser()

        parser.add_subcommands(CmdA())
        parser.add_subcommands(CmdB(), CmdC())

        cmd = 'cmda'
        argsA = parser.parse_args([cmd])
        parser.exec_subcommands(argsA)

        cmd = 'cmdb'
        argsB = parser.parse_args([cmd])
        parser.exec_subcommands(argsB)

        cmd = 'cmdc'
        argsC = parser.parse_args([cmd])
        parser.exec_subcommands(argsC)

        cmd = 'cmda'
        parser.exec_subcommands(argsA)

    def test_argcomplete(self):
        cmd = ''

        class CmdA(Subcommand):
            def on_parser_init(self, parser: argparse.ArgumentParser):
                pass

            def on_command(self, args):
                assert cmd == 'cmda'

        parser = SubcommandParser(argcomplete=True)

        parser.add_subcommands(CmdA())

        cmd = 'cmda'
        args = parser.parse_args([cmd])
        parser.exec_subcommands(args)

    def test_help(self):
        cmd = ''

        class CmdA(Subcommand):
            def on_parser_init(self, parser: argparse.ArgumentParser):
                pass

            def on_command(self, args):
                pass

        parser = SubcommandParser()

        parser.add_subcommands(CmdA())

        cmd = '--help'

        try:
            args = parser.parse_args([cmd])
        except:
            pass

    def test_shared_argument(self):
        cmd = ''

        class CmdA(Subcommand):
            def on_parser_init(self, parser: argparse.ArgumentParser):
                pass

            def on_command(self, args):
                assert cmd == 'cmda'

        parser = SubcommandParser()

        parser.add_argument('-x', action='store_true', shared=True)

        parser.add_subcommands(CmdA())

        # parser.add_argument('-y', action='store_true', shared=True)

        cmd = 'cmda'
        # args = parser.parse_args([cmd, '-x', '-y'])
        args = parser.parse_args([cmd, '-x'])
        parser.exec_subcommands(args)

    def test_subsubcommand(self):
        cmd = ''

        class SubSubCmd(Subcommand):
            def on_parser_init(self, parser: SubcommandParser):
                parser.add_argument('x', type=str)

            def on_command(self, args):
                assert args.x == 'arg'

        class SubCmd(Subcommand):
            def on_parser_init(self, parser: SubcommandParser):
                parser.add_subcommands(SubSubCmd())

            def on_command(self, args):
                pass

        parser = SubcommandParser()

        parser.add_subcommands(SubCmd())

        cmd = 'subcmd subsubcmd arg'.split()
        args = parser.parse_args(*[cmd])
        parser.exec_subcommands(args)

    def test_unknown_args(self):
        class SubCmdA(Subcommand):
            def on_parser_init(self, parser: SubcommandParser):
                parser.add_argument('foo')

            def on_command(self, args):
                assert args.foo == 'foo1'

        class SubCmdB(Subcommand):
            def on_parser_init(self, parser: SubcommandParser):
                parser.allow_unknown_args = True
                parser.add_argument('foo')

            def on_command(self, args, unknown_args=None):
                assert args.foo == 'foo1'
                assert isinstance(unknown_args, list)
                assert len(unknown_args) == 2
                assert unknown_args[0] == 'unknown'
                assert unknown_args[1] == 'args'

        parser = SubcommandParser()
        # parser.allow_unknown_args = True

        parser.add_subcommands(SubCmdA())
        parser.add_subcommands(SubCmdB())

        argv = sys.argv.copy()

        sys.argv[1:] = ['subcmda', 'foo1']

        parser.exec_subcommands()

        sys.argv[1:] = ['subcmda', 'foo1', 'unknown', 'args']

        try:
            parser.exec_subcommands()
        except:
            pass
        else:
            raise Exception('did not raise exception for unknown args')

        sys.argv[1:] = ['subcmdb', 'foo1', 'unknown', 'args']

        parser.exec_subcommands()

        sys.argv = argv



from unittest import TestCase, SkipTest

import yautil
import matplotlib.pyplot as plt


class TestPlot(TestCase):
    plot: callable
    pause = .001

    @classmethod
    def setUpClass(cls):
        if cls is TestPlot:
            raise SkipTest("Skip BaseTest tests, it's a base class")
        super(TestPlot, cls).setUpClass()

    def test_basic(self):
        data = [0, 1, 2, 2, 3, 4]
        self.plot(data, block=False)
        plt.pause(self.pause)
        plt.close()

    def test_xlabel(self):
        data = ['test values', 0, 1, 2, 2, 3, 4]
        self.plot(data, block=False)
        plt.pause(self.pause)
        plt.close()

    def test_subfigures(self):
        data = [0, 1, 2, 2, 3, 4]
        self.plot(data, data, block=False)
        plt.pause(self.pause)
        plt.close()

    def test_multi_lines(self):
        data1 = ['test values 1', 0, 1, 2, 2, 3, 4]
        data2 = ['test values 2', 1, 2, 2, 3, 4]
        self.plot((data1, data2), block=False)
        plt.pause(self.pause)
        plt.close()


class TestPlotCdf(TestPlot):

    def __init__(self, *args, **kwargs):
        self.plot = yautil.plot_cdf
        super().__init__(*args, **kwargs)


class TestPlotLinear(TestPlot):

    def __init__(self, *args, **kwargs):
        self.plot = yautil.plot_linear
        super().__init__(*args, **kwargs)


class TestPlotScatter(TestPlot):

    def __init__(self, *args, **kwargs):
        self.plot = yautil.plot_scatter
        super().__init__(*args, **kwargs)


class TestPlotBox(TestPlot):

    def __init__(self, *args, **kwargs):
        self.plot = yautil.plot_box
        super().__init__(*args, **kwargs)



import os
from unittest import TestCase

from yautil import PersistentCache
from tempfile import TemporaryDirectory


class TestCache(TestCase):
    tmpdir: TemporaryDirectory
    cache_dir: str
    files_dir: str

    def setUp(self):
        self.tmpdir = TemporaryDirectory()
        self.cache_dir = os.path.join(self.tmpdir.name, 'cache')
        os.mkdir(self.cache_dir)
        self.files_dir = os.path.join(self.tmpdir.name, 'files')
        os.mkdir(self.files_dir)

    def tearDown(self) -> None:
        self.tmpdir.cleanup()

    def test_rw(self):
        pc = PersistentCache(cache_dir=self.cache_dir)

        key = pc.StrKey('x')

        assert key not in pc

        pc.update(key, 'val_x')

        assert key in pc

        value = pc.lookup(pc.StrKey('x'))

        assert value == 'val_x'

    def test_rw_filekey(self):
        pc = PersistentCache(cache_dir=self.cache_dir)

        f1 = os.path.join(self.files_dir, 'f1')
        with open(f1, 'w') as f:
            f.write('hello')

        key = pc.FileKey(f1)

        assert key not in pc

        pc.update(key, 'val_1')

        assert key in pc

        with open(f1, 'w') as f:
            f.write('bye')

        key2 = pc.FileKey(f1)

        assert key2 not in pc

        assert pc.lookup(key) == 'val_1'

        pc.update(key2, 'val_2')

        assert key in pc
        assert key2 in pc

        pc.remove(key)

        assert key not in pc
        assert key2 in pc

        assert pc.lookup(key2) == 'val_2'



# PYTHON_ARGCOMPLETE_OK

import argparse
from typing import Union, List
from gettext import gettext as _

from . import DummyModule
from .pyshutil import compile_shargs

try:
    import argcomplete
except ImportError:
    pass


class SubcommandParser(argparse.ArgumentParser):
    subparsers = None
    shared_parser = None

    argcomplete: bool
    __allow_unknown_args: bool

    __unknown_args: list = None

    @property
    def allow_unknown_args(self) -> bool:
        return self.__allow_unknown_args

    @allow_unknown_args.setter
    def allow_unknown_args(self, val: bool):
        self.__allow_unknown_args = val

    def __init__(self, *args, argcomplete: bool = False, **kwargs):
        super().__init__(*args, **kwargs)

        self.argcomplete = argcomplete
        self.allow_unknown_args = False

    def add_subcommands(self, *subcommands, title='subcommands', required=True, help=None, metavar=None):
        if not self.subparsers:
            self.subparsers = self.add_subparsers(
                title=title,
                required=required,
                help=help,
                metavar=metavar,
            )
            self.subparsers.dest = 'subcommand'

        for subcommand in subcommands:
            if not isinstance(subcommand, Subcommand):
                raise TypeError(str(subcommand.__class__) + 'is not Subcommand')
            subcommand._register(self.subparsers,
                                 parent=self.shared_parser)

    def try_argcomplete(self):
        if 'argcomplete' in globals() and not isinstance(argcomplete, DummyModule):
            argcomplete.autocomplete(self)
        else:
            print('warning: install \'argcomplete\' package to enable bash autocomplete')

    def parse_args(self, *args, **kwargs) -> any:
        if self.argcomplete:
            self.try_argcomplete()
        parsed_args, unknown_args = super().parse_known_args(*args, **kwargs)
        if unknown_args and not parsed_args._allow_unknown_args:
            msg = _('unrecognized arguments: %s')
            self.error(msg % ' '.join(unknown_args))
        self.__unknown_args = unknown_args
        return parsed_args

    def exec_subcommands(self, parsed_args: object = None):
        if not parsed_args:
            parsed_args = self.parse_args()

        if parsed_args._allow_unknown_args:
            parsed_args._func(parsed_args, unknown_args=self.__unknown_args)
        else:
            parsed_args._func(parsed_args)

    def add_argument(self, *args, shared: bool = False, **kwargs):
        if shared:
            if not self.shared_parser:
                self.shared_parser = argparse.ArgumentParser(add_help=False)

            return self.shared_parser.add_argument(*args, **kwargs)

        return super().add_argument(*args, **kwargs)


class Subcommand:
    parser: SubcommandParser
    name: str
    help: str

    def on_parser_init(self, parser: SubcommandParser):
        raise NotImplementedError

    def on_command(self, args, unknown_args=None):
        raise NotImplementedError

    def _register(self, subparsers, parent: argparse.ArgumentParser = None):
        kwargs = {'help': self.help}
        if parent:
            kwargs['parents'] = [parent]

        self.parser = subparsers.add_parser(self.name, **kwargs)
        self.parser.__class__ = SubcommandParser
        self.on_parser_init(self.parser)
        self.parser.set_defaults(
            _func=self.on_command,
            _allow_unknown_args=self.parser.allow_unknown_args,
        )
        subparsers.metavar = 'command'

    def __init__(self, subparsers = None, name: str = None, help: str = '', dependency: Union[str, List[str]] = ''):
        self.name = name if name else type(self).__name__.lower()
        self.help = help
        if subparsers:
            self._register(subparsers)

    @classmethod
    def exec(cls, *args, **kwargs):
        cmdargs, shargs = compile_shargs(*args, **kwargs)

        parser = SubcommandParser()
        parser.add_subcommands(cls(name='subcmd'))
        parserd_args = parser.parse_args(['subcmd', *cmdargs])

        parser.exec_subcommands(parserd_args)



import multiprocessing as mp
import signal
import sys
import traceback
import uuid
from typing import List, Tuple

from tqdm import tqdm

_obj_idmap: dict = {}


# https://gist.github.com/EdwinChan/3c13d3a746bb3ec5082f
# globalize decorator for inner functions
def globalize(func):
    def result(*args, **kwargs):
        return func(*args, **kwargs)

    result.__name__ = result.__qualname__ = uuid.uuid4().hex
    setattr(sys.modules[result.__module__], result.__name__, result)
    return result


# Shortcut to multiprocessing's logger
def error(msg, *args):
    return mp.get_logger().error(msg, *args)


class LogExceptions(object):
    def __init__(self, callable):
        self.__callable = callable

    def __call__(self, *args, **kwargs):
        try:
            result = self.__callable(*args, **kwargs)

        except Exception as e:
            # Here we add some debugging help. If multiprocessing's
            # debugging is on, it will arrange to log the traceback
            error(traceback.format_exc())
            # Re-raise the original exception so the Pool worker can
            # clean up
            raise

        # It was fine, give a normal answer
        return result


def _sighandler(signum, frame):
    if signum == signal.SIGALRM:
        signal.signal(signal.SIGALRM, _sighandler)
        raise TimeoutError('MpUtil: Timeout expired')
    else:
        raise Exception('MpUtil: Unexpected signal received: ' + str(signum))


def _on_done(result):
    self_id, count, ret = result
    # print('done: self_id: ' + str(self_id) + ', count: ' + str(count) + ', res: ' + str(ret))

    self = _obj_idmap[self_id]

    with self.free.get_lock():
        self.results.extend(ret)
        while ret:
            ret.pop()
        self.buffer.append(ret)
        if self.pbar:
            self.pbar.update(count)

        self.free.value += 1
        self.free_cond.notify()


def _func_wrapper(self_id, ret: list, arglist: List[Tuple]):
    count: int = 0
    # print('wrapper: id: ' + str(self_id) + ', ret: ' + str(ret))
    self = _obj_idmap[self_id]

    if self.task_timeout > 0:
        signal.signal(signal.SIGALRM, _sighandler)

    for argset in arglist:
        func, args, kwargs, _siz = argset
        # print('wrapper: self_id: ' + str(self_id) + ', func: ' + str(func) + ', args: ' + str(args) + ', kwargs: ' + str(kwargs))
        if self.task_timeout > 0:
            signal.alarm(self.task_timeout)
        try:
            if self.arg0:
                result = func(*args, arg0=self.arg0, **kwargs)
            else:
                result = func(*args, **kwargs)
            if self.task_timeout > 0:
                signal.alarm(0)
        except Exception as e:
            traceback.print_tb(e.__traceback__)
            print('MpUtil error: ' + str(e) + ' while processing *args: ' + str(args) + ', **kwargs: ' + str(kwargs))
            result = None

        if result is not None:
            try:
                ret.append(result)
            except Exception as e:
                print('MpUtil error: can\'t handle results from ' + str(func) + '(' + str(args) + ', ' + str(kwargs) + '). ' + str(e))
        count += _siz

    return self_id, count, ret


class MpUtil:
    mpctx: mp
    ctx: mp.Manager
    pool: mp.Pool
    pbar: tqdm
    free = None
    queue: list
    results: list
    buffer: List[List]
    id = None
    maxchunksize: int
    task_timeout: int
    arg0: object

    def __init__(self, processes: int = 0, maxchunksize: int = 1, total: int = 0, desc: str = "", task_timeout=0,
                 arg0: object = None, pbar: bool = True):
        if processes == 0:
            processes = mp.cpu_count()

        if 'fork' in mp.get_all_start_methods():
            self.mpctx = mp.get_context('fork')
        else:
            raise OSError('\'fork\' starting method is required from the multiprocessing module')

        self.ctx = self.mpctx.Manager()
        self.free = self.mpctx.Value('i', processes)
        self.free_cond = self.mpctx.Condition(self.free.get_lock())

        self.queue = []
        self.results = self.ctx.list([])
        self.pbar = tqdm(total=total, desc=desc) if pbar else None

        self.buffer = self.ctx.list()
        for _ in range(processes + 1):
            self.buffer.append(self.ctx.list())

        self.maxchunksize = maxchunksize
        self.task_timeout = task_timeout
        self.arg0 = arg0

        self.id = id(self)
        _obj_idmap[self.id] = self

        self.pool = self.mpctx.Pool(processes=processes)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __del__(self):
        if _obj_idmap:
            _obj_idmap.pop(self.id)

    def flush(self):
        ret: list = self.buffer.pop()
        self.pool.apply_async(LogExceptions(_func_wrapper), (self.id, ret, self.queue,), callback=_on_done)
        self.queue = []
        self.free.value -= 1

    def schedule(self, func: callable, *args, _siz: int = 1, **kwargs):
        with self.free.get_lock():
            self.queue.append((func, args, kwargs, _siz))

            if self.free.value > 0:
                self.flush()
                return

            if len(self.queue) < self.maxchunksize:
                return

            while self.free.value == 0:
                self.free_cond.wait()
            self.flush()

    def wait(self):
        self.flush()
        self.pool.close()
        self.pool.join()
        if self.pbar:
            self.pbar.close()
        self.mpctx.log_to_stderr()
        return self.results



from enum import Enum, auto
from typing import Union, List, Tuple, Dict


class __PlotMode(Enum):
    LINEAR = auto()
    SCATTER = auto()
    BOX = auto()


def __unzip_subplot_data(subplot_data: Tuple) -> (list, list):
    labels = []
    lines = []
    for line_data in subplot_data:
        assert isinstance(line_data, list)

        if line_data and isinstance(line_data[0], str):
            line_label = line_data[0]
            line_data = line_data[1:]
        else:
            line_label = None

        if not line_data:
            continue

        labels.append(line_label)
        lines.append(line_data)

    return labels, lines


def __plot(*data,
           titles: Union[str, List[str]] = None,
           maxcol: int = 4,
           xlabel: str = None,
           ylabel: str = None,
           xlim: List = None,
           ylim: List = None,
           padding: float = 0.2,
           block: bool = True,
           legend_loc: str = None,
           font_family: str = None,
           autofmt_xdate: bool = True,
           # the below are private params
           mode: __PlotMode = __PlotMode.LINEAR,
           figsize: tuple = None,
           subplots_adjust: Dict[str, float] = None,
           **subplot_kwargs
           ):
    """
    Usage examples:

    * Data should be a list of integers

      plot_cdf( [0, 1, 2, 3, 4] )

    * A string element at the beginning of a list is taken as data label.

      plot_cdf( ['example seq.', 0, 1, 2, 3, 4] )

    * Multiple data are plotted on different subfigures.

      plot_cdf( [0, 1, 2, 3, 4], [2, 4, 8, 6, 0] )

    * Data in the same tuple are plotted on the same subfigure.

      plot_cdf( ([0, 1, 2, 3, 4], [2, 4, 8, 6, 0]) )


    :param data: List(s) of integers
    :param titles: Title(s) of each subfigure
    :param maxcol: Maximum number of subfigures in the same row
    :param xlabel: Label (e.g., 'sec')
    :param padding:
    """

    import matplotlib.pyplot as plt

    if not titles:
        titles = []

    fig = plt.figure(figsize=figsize)
    plt.clf()
    if font_family:
        # 'NanumGothicOTF'
        plt.rc('font', family=font_family)

    for i, subplot_data in enumerate(data):
        if not isinstance(subplot_data, tuple):
            assert isinstance(subplot_data, list)
            subplot_data = (subplot_data,)

        subplot = fig.add_subplot(int((len(data) - 1) / maxcol + 1),
                                  int(min(len(data), maxcol)),
                                  int(i + 1))
        if i < len(titles):
            subplot.title.set_text(titles[i])
        if xlabel:
            subplot.set_xlabel(xlabel)
        if ylabel:
            subplot.set_ylabel(ylabel)
        if xlim:
            subplot.set_xlim(xlim)
        if ylim:
            subplot.set_ylim(ylim)
        plt.subplots_adjust(left=padding, right=1 - padding, top=1 - padding, bottom=padding)

        labels, lines = __unzip_subplot_data(subplot_data)

        if mode == __PlotMode.SCATTER or mode == __PlotMode.LINEAR:
            for label, line in zip(labels, lines):
                assert isinstance(line[0], tuple)

                x, y = map(list, zip(*line))

                args, kwargs = (lambda *a, **ka: (a, ka))(x, y, label=label, **subplot_kwargs)

                if mode == __PlotMode.SCATTER:
                    subplot.scatter(*args, **kwargs)
                elif mode == __PlotMode.LINEAR:
                    subplot.plot(*args, **kwargs)
        elif mode == __PlotMode.BOX:
            subplot.boxplot(lines, labels=labels, **subplot_kwargs)

    if legend_loc:
        plt.legend(loc=legend_loc)

    if autofmt_xdate:
        fig.autofmt_xdate()

    if subplots_adjust:
        plt.subplots_adjust(**subplots_adjust)

    plt.show(block=block)


def __modify_line_data(data, modify_line_data: callable) -> list:
    modified_data = []

    for subplot_data in data:
        if not isinstance(subplot_data, tuple):
            subplot_data = (subplot_data,)

        modified_subplot_data = []

        for line_data in subplot_data:
            assert isinstance(line_data, list)

            if line_data and isinstance(line_data[0], str):
                line_label = line_data[0]
                line_data = line_data[1:]
            else:
                line_label = None

            if line_label:
                modified_line_data = [line_label]
            else:
                modified_line_data = []

            modified_line_data.extend(modify_line_data(line_data))

            modified_subplot_data.append(modified_line_data)

        modified_data.append(tuple(modified_subplot_data))

    return modified_data


def plot_cdf(*data,
             titles: Union[str, List[str]] = None,
             maxcol: int = 4,
             xlabel: str = None,
             padding: float = 0.2,
             block: bool = True,
             legend_loc: str = None,
             font_family: str = None,
             figsize: tuple = None,
             subplots_adjust: Dict[str, float] = None,
             **subplot_kwargs
             ):
    """
    Usage examples:

    * Data should be a list of integers

      plot_cdf( [0, 1, 2, 3, 4] )

    * A string element at the beginning of a list is taken as data label.

      plot_cdf( ['example seq.', 0, 1, 2, 3, 4] )

    * Multiple data are plotted on different subfigures.

      plot_cdf( [0, 1, 2, 3, 4], [2, 4, 8, 6, 0] )

    * Data in the same tuple are plotted on the same subfigure.

      plot_cdf( ([0, 1, 2, 3, 4], [2, 4, 8, 6, 0]) )


    :param font_family:
    :param data: List(s) of integers
    :param titles: Title(s) of each subfigure
    :param maxcol: Maximum number of subfigures in the same row
    :param xlabel: Label (e.g., 'sec')
    :param padding:
    :param block:
    :param legend_loc: One of
        'best', 'upper right', 'upper left', 'lower left',
        'lower right', 'right', 'center left', 'center right',
        'lower center', 'upper center', and 'center'
    """

    def __to_cdf(line_data: list) -> list:
        import numpy as np

        # sort the data:
        d = np.sort(line_data)

        # calculate the proportional values of samples
        p = 1. * np.arange(len(line_data)) / (len(line_data) - 1)

        return list(zip(d, p))

    data = __modify_line_data(data, __to_cdf)

    __plot(*data,
           titles=titles,
           maxcol=maxcol,
           xlabel=xlabel,
           ylabel='CDF',
           ylim=[0,1],
           padding=padding,
           block=block,
           legend_loc=legend_loc,
           font_family=font_family,
           figsize=figsize,
           subplots_adjust=subplots_adjust,
           **subplot_kwargs
           )


def plot_linear(*data,
                titles: Union[str, List[str]] = None,
                maxcol: int = 4,
                xlabel: str = None,
                ylabel: str = None,
                xlim: List = None,
                ylim: List = None,
                padding: float = 0.2,
                block: bool = True,
                legend_loc: str = None,
                font_family: str = None,
                figsize: tuple = None,
                subplots_adjust: Dict[str, float] = None,
                **subplot_kwargs
                ):
    """
    Usage examples:

    * Data should be a list of integers

      plot_cdf( [0, 1, 2, 3, 4] )

    * A string element at the beginning of a list is taken as data label.

      plot_cdf( ['example seq.', 0, 1, 2, 3, 4] )

    * Multiple data are plotted on different subfigures.

      plot_cdf( [0, 1, 2, 3, 4], [2, 4, 8, 6, 0] )

    * Data in the same tuple are plotted on the same subfigure.

      plot_cdf( ([0, 1, 2, 3, 4], [2, 4, 8, 6, 0]) )


    :param font_family:
    :param data: List(s) of integers
    :param titles: Title(s) of each subfigure
    :param maxcol: Maximum number of subfigures in the same row
    :param xlabel: Label (e.g., 'sec')
    :param ylabel:
    :param xlim:
    :param ylim:
    :param padding:
    :param block:
    :param legend_loc: One of
        'best', 'upper right', 'upper left', 'lower left',
        'lower right', 'right', 'center left', 'center right',
        'lower center', 'upper center', and 'center'
    """

    def __add_x_values_if_necessarily(line_data: list) -> list:
        if isinstance(line_data[0], tuple):
            return line_data

        return list(zip(range(len(line_data)), line_data))

    data = __modify_line_data(data, __add_x_values_if_necessarily)

    __plot(*data,
           titles=titles,
           maxcol=maxcol,
           xlabel=xlabel,
           ylabel=ylabel,
           xlim=xlim,
           ylim=ylim,
           padding=padding,
           block=block,
           legend_loc=legend_loc,
           font_family=font_family,
           figsize=figsize,
           subplots_adjust=subplots_adjust,
           **subplot_kwargs
           )


def plot_scatter(*data,
                 titles: Union[str, List[str]] = None,
                 maxcol: int = 4,
                 xlabel: str = None,
                 ylabel: str = None,
                 xlim: List = None,
                 ylim: List = None,
                 padding: float = 0.2,
                 block: bool = True,
                 legend_loc: str = None,
                 font_family: str = None,
                 figsize: tuple = None,
                 subplots_adjust: Dict[str, float] = None,
                 **subplot_kwargs
                 ):
    """
    Usage examples:

    * Data should be a list of integers

      plot_cdf( [0, 1, 2, 3, 4] )

    * A string element at the beginning of a list is taken as data label.

      plot_cdf( ['example seq.', 0, 1, 2, 3, 4] )

    * Multiple data are plotted on different subfigures.

      plot_cdf( [0, 1, 2, 3, 4], [2, 4, 8, 6, 0] )

    * Data in the same tuple are plotted on the same subfigure.

      plot_cdf( ([0, 1, 2, 3, 4], [2, 4, 8, 6, 0]) )


    :param font_family:
    :param data: List(s) of integers
    :param titles: Title(s) of each subfigure
    :param maxcol: Maximum number of subfigures in the same row
    :param xlabel: Label (e.g., 'sec')
    :param ylabel:
    :param xlim:
    :param ylim:
    :param padding:
    :param block:
    :param legend_loc: One of
        'best', 'upper right', 'upper left', 'lower left',
        'lower right', 'right', 'center left', 'center right',
        'lower center', 'upper center', and 'center'
    """

    def __add_x_values_if_necessarily(line_data: list) -> list:
        if isinstance(line_data[0], tuple):
            return line_data

        return list(zip(range(len(line_data)), line_data))

    data = __modify_line_data(data, __add_x_values_if_necessarily)

    __plot(*data,
           titles=titles,
           maxcol=maxcol,
           xlabel=xlabel,
           ylabel=ylabel,
           xlim=xlim,
           ylim=ylim,
           padding=padding,
           block=block,
           legend_loc=legend_loc,
           font_family=font_family,
           mode=__PlotMode.SCATTER,
           figsize=figsize,
           subplots_adjust=subplots_adjust,
           **subplot_kwargs
           )


def plot_box(*data,
             titles: Union[str, List[str]] = None,
             maxcol: int = 4,
             xlabel: str = None,
             ylabel: str = None,
             ylim: List = None,
             padding: float = 0.2,
             block: bool = True,
             legend_loc: str = None,
             font_family: str = None,
             figsize: tuple = None,
             subplots_adjust: Dict[str, float] = None,
             **subplot_kwargs
             ):
    __plot(*data,
           titles=titles,
           maxcol=maxcol,
           xlabel=xlabel,
           ylabel=ylabel,
           ylim=ylim,
           padding=padding,
           block=block,
           legend_loc=legend_loc,
           font_family=font_family,
           mode=__PlotMode.BOX,
           figsize=figsize,
           subplots_adjust=subplots_adjust,
           **subplot_kwargs
           )
